#!/usr/bin/perl
#  2018-05-27  Tim Godfrey
# Generate reports in PCGtools.
# Patch Report (Patch is default report setting.  select CSV output file, fix output path to C:\Data\Music\Kronos\Backup
# set output name t_180630_patch_output.csv
# Usage Report (setting List Type: select Usage, select CSV output file, fix output path to C:\Data\Music\Kronos\Backup
# set output name t_180630_usage_output.csv
#
# Usage:
# perl Kronos_Used_Unused_Programs.pl  <Program Usage Report.csv> <Patch List Report.csv> 

use File::Basename;
use Data::Dumper;

use strict;
use warnings;

#+=======+=====+=====================================================+==============+====================+=====+===+===+===+===+
#
#Note: 
my $Usagefile;
my $Patchfile;
my $Setlistfile;
my @combilines;
my $work;
my @ProgLine;  # use split to separate CSV into this array
my $ThisProgID;
my $linecnt = 0;
my %CombiHash;  # Gives name of Combi for Combi-ID as Key
my %SetListUsageHash;  # Combi-ID is Key, value provides a comma delimited list of Setlist ID's using this Combi
#my %SetListNamesHash;  # Gives Setlist bank/slot names with setlistID as Key
my %SetlistHash;  # Gives name of Setlist for numeric setlist-ID as Key
my %UsageHash;   #  Give comma-delimiated string of Combi-ID's Using this Prog-ID
my %ProgIsUsedHash;   # Value is * for Prog ID's that are used
my %CombiIsUsedHash;   # Value is * for Combi ID's that are used
my $UsedProgOutFileName;
my $longUsedProgOutFileName;
my $unUsedProgOutFileName;
my $AsciiOutFilename;
my $UsedProgOutFile;
my $longUsedProgOutFile;
my $unUsedProgOutFile;
my $AsciiOutFile;
my $UsedCombiOutFileName;
my $longUsedCombiOutFileName;
my $unUsedCombiOutFileName;
my $UsedCombiOutFile;
my $longUsedCombiOutFile;
my $unUsedCombiOutFile;
my $printline;

# Read through Usage list and save to Hash %UsageHash
# format of CSV line:
# I-A003,I-A030,I-B104,I-D017,U-E000,U-F004,U-G004,U-G014,U-G015,
# Combi   prg1   prg2   prg3   etc
#
#
if (defined ($ARGV[0]) )  {    # use first filename for the Programs Usage list (.CSV format)
   print "Opening $ARGV[0] for reading\n";
   open ($Usagefile, '<', $ARGV[0])
     or die "Could not open $ARGV[0]\n";

   while (my $work = <$Usagefile>) {
     chomp $work;
     $work =~ s/^\s+//;    # remove leading whitespace from work line
     
     my($first, $rest) = split(/,/, $work, 2);
     $first =~ s/^\s+|\s+$//g;  # trim leading/trailing
     $rest =~ s/^\s+|\s+$//g;  # trim leading/trailing
#     print("$first ||  $rest \n"); 
     $UsageHash{$first} = $rest;   # set Combi number ID as Key and Name as Value
     
     
     } 
   close ($Usagefile);

 } else {
      die "PCG Tools Programs Usage report list (.CSV format) Filename Required as 1st parameter \n (Ignore Empty/Init Combi/Prog should be checked)\nPCG Tools Patch list (.CSV format) Filename Required as 2nd parameter\n Die";
 }


# ---------------- Build Hash of Combi Names --------------------------------------------------------------------
#  Read through patch list and save all combi data to hash %CombiHash  as  ID / Name pair
# format of Combi patch list .csv
# KRONOS German Grand,Program,I-A000, ,,Keyboard,A.Piano,,,
#  [0]                  [1]     [2]

if (defined($ARGV[1]))  {    # use second filename for the All-Patch list (.CSV format)
   print "Opening $ARGV[1] for reading\n";
   open ($Patchfile, '<', $ARGV[1])
     or die "Could not open $ARGV[1]\n";

   while (my $work = <$Patchfile>) {
     chomp $work;
     $work =~ s/^\s+//;    # remove leading whitespace from work line
     
     if (index($work, ",Combi  ") != -1) {        #only pay attention to COMBI lines
        $linecnt = $linecnt + 1;
        my @CombiLine = split(',', $work);   # parse on commas for CSV
        my $CombiID = sprintf("%-9s",$CombiLine[4]);  # Combi ID is 5th element of array after split [4]
#        print "$work| Making CombiHash: |$CombiLine[2]|\n";                       #       
#        $CombiHash{$CombiLine[2]} = $CombiID.$CombiLine[0];   # set Combi number ID as Key and Name as Value
        $CombiHash{$CombiLine[2]} = $CombiLine[0];   # set Combi number ID as Key and Name as Value
        
#        print "CombiHash - key $CombiLine[4]    ~~~~   val: $CombiID.$CombiLine[0]   \n";
     }
     
                                                        #                                                   [0]        [1]     [2]     [6]
     if (index($work, ",SetListSlot,") != -1) {        #Now pay attention to SetListSlot lines format:  Oceans 1,SetListSlot,000/58,,,,Colonial 1,
#        $work = substr($work, 1);  # drop first character, an extra comma
        my @SetlLine = split(',', $work);
#        my $SetlID = substr($SetlLine[2],0,3);   # just the 3 digit Setlist bank number
        my $SetlID = $SetlLine[2];   # The whole Setlist bank ID
#        print "$SetlID    >>>>>>>>    ";
        $SetlistHash{$SetlID} = $SetlLine[2]." ".$SetlLine[6]." / ".$SetlLine[0];   # set Setlist bank number ID as Key and Name as Value
        
#        print "$SetlistHash{$SetlID}   \n";
     }
  } 
  close ($Patchfile);
 } else {
      die "PCG Tools Patch list (.CSV format) Filename Required as 2nd parameter";
 }










# ---------------- Build Hash of Combi's used by Setlist Slots --------------------------------------------------------------------

#  Read through setlist data list and save those that refer to combis
# format of Setlist dat list .csv
# 009/12, Set List 009 TESTS, ,    Combi,   I-A000, Color 12, Font 0,  Transpose 0, HoldTime 6, Volume 127, Kbd_Track 0
#  [0]           [1]        [2]      [3]     [4]
# SL-ID      SL bankname    SLslot  Type    CombiID 

if (defined($ARGV[2]))  {    # use third filename for the setlist output (.CSV format)
   print "Opening $ARGV[2] for reading\n";
   open ($Setlistfile, '<', $ARGV[2])
     or die "Could not open $ARGV[2]\n";

   while (my $work = <$Setlistfile>) {
     chomp $work;
     $work =~ s/^\s+//;    # remove leading whitespace from work line
     
     if (index($work, ", Combi,") != -1) {        #only pay attention to COMBI lines
        $linecnt = $linecnt + 1;
#        print "$work\n";
        my @CombiLine = split(',', $work);
        my $CombiID = $CombiLine[4];
        $CombiID  =~ s/^\s+|\s+$//g;    # trim lead/trail spaces
#        print "|$CombiID|\n";
        
        if ($SetListUsageHash{$CombiID})  {    # is this Combi number ID in the SetlistUsagehash? 
            $SetListUsageHash{$CombiID} = $SetListUsageHash{$CombiID}.",".$CombiLine[0];   # Append Combi number ID to Key string of comma separated Setlist bank/slots in  Value
#            print "Append SetListUsageHash - key |$CombiID|    ~~~~   val: $SetListUsageHash{$CombiID}   \n";
            
            }  else {
            $SetListUsageHash{$CombiID} = $CombiLine[0];   # set Combi number ID as Key string of comma separted Setlist bank/slots in  Value
#            print "New   SetListUsageHash - key |$CombiID|    ~~~~   val: $SetListUsageHash{$CombiID}    \n";
        }
     }
 
  } 
  close ($Setlistfile);
 } else {
      die "Setlist output from pcgparsesetlist.exe (.CSV format) Filename Required as 3rd parameter";
 }











$linecnt = 0;

# ---------------Output files for programs ------------------

$UsedProgOutFileName = $ARGV[0];
$UsedProgOutFileName =~ s/\..*$/.usedprog.txt/;  # force output filename to .usedprog.txt
print "opening $UsedProgOutFileName for writing\n";
open ($UsedProgOutFile, '>'.$UsedProgOutFileName);    

print $UsedProgOutFile "Used Program List for $ARGV[0] and $ARGV[1]\n";
print $UsedProgOutFile "--------------------------------------------------------------\n";

$longUsedProgOutFileName = $UsedProgOutFileName;  
$longUsedProgOutFileName =~ s/\..*$/.longusedprog.txt/;  # force output filename to .longusedprog.txt
print "opening $longUsedProgOutFileName for writing\n";
open ($longUsedProgOutFile, '>'.$longUsedProgOutFileName);    
 
print $longUsedProgOutFile "Long Format (with referencing Combi ID and Names) Used Program List for $ARGV[0]\n";
print $longUsedProgOutFile "--------------------------------------------------------------\n";


$unUsedProgOutFileName = $ARGV[0];
$unUsedProgOutFileName =~ s/\..*$/.unusedprog.txt/;  # force output filename to .unusedprog.txt
print "opening $unUsedProgOutFileName for writing\n";
open ($unUsedProgOutFile, '>'.$unUsedProgOutFileName);    
print $unUsedProgOutFile "Unused Program List for $ARGV[0] and $ARGV[1]\n";
print $unUsedProgOutFile "--------------------------------------------------------------\n";





#  Read through patch list again and look at programs 
#     Look up 
# format of Program in Patch List
# KRONOS German Grand,Program,I-A000, ,,Keyboard,A.Piano,,,
#  [0]                  [1]     [2]   

if (defined($ARGV[1]))  {    # use second filename for the Patch list (.CSV format)
   print "Opening $ARGV[1] for reading\n";
   open (my $Patchfile, '<', $ARGV[1])
     or die "Could not open $ARGV[1]\n";

   while (my $work = <$Patchfile>) {
     chomp $work;
     $work =~ s/^\s+//;    # remove leading whitespace from work line
     
     if (index($work, ",Program,") != -1) {    #is the patch list line referencing  a program? 
        $linecnt = $linecnt + 1;
        @ProgLine = split(',', $work);   # split patch list line on commas
        
        # format of line:  Acoustic Bass 2 SW1,Program,I-D097, ,,Bass/Synth Bass,A.Bass,,,
        #                        [0]              [1]     [2]
        #
        my $ThisProgID = $ProgLine[2];
        $ThisProgID =~ s/^\s+|\s+$//g;  #trim lead/trail spaces
        if ($UsageHash{$ThisProgID})  {    # it this program in the patch list found in Usage List hash? 
               my $CombisUsing = $UsageHash{$ThisProgID};    # get the string of Combi IDs that use this program
               $printline  = sprintf("%-7s : %-24s UsedBy: ",$ThisProgID, $ProgLine[0]);   # print the start of the short format used list
#               print "$printline";
               print $UsedProgOutFile "$printline";
               my $UseArrayLine = sprintf("%-7s *",$ThisProgID);   
               $ProgIsUsedHash{$ThisProgID} = "*";      #  record used program with *
               
               #  Now write out the long format with Using Combi Names
               my(@CombisUsingLine) = split(/,/, $CombisUsing);    # Parse comma delimited items in the string into an array
               $printline  = sprintf("%-7s : %-24s",$ThisProgID, $ProgLine[0]);  # format the line to print
               print $longUsedProgOutFile "$printline\n";

               foreach (@CombisUsingLine)  {    
                   my $itemID = $_;
#                   print "$itemID\n";                  #                                ****************************************
                   $itemID =~ s/^\s+|\s+$//g;  #trim lead/trail spaces   
                   if (defined($CombiHash{$itemID})) {       # Is it a known combi from the patch list? 
                      if (index($CombiHash{$itemID}, "Init Combi") == -1) {  # and it is not an init combi
                           print $UsedProgOutFile "$itemID,";    # print only the Combi ID in short used file
                           print $longUsedProgOutFile "          $itemID $CombiHash{$itemID}\n";    #  Print out the Combi ID and Name for long list
                        }
                   }
                else { 
                    if (defined($SetlistHash{$itemID})) {    # is it a known setlist from the patch list? 
                       print $UsedProgOutFile "$itemID,";    # print only the setlist ID in short used file
                       print $longUsedProgOutFile "          Setlist $SetlistHash{$itemID} \n";    # print out the setlist bank and name
                    } else {                                               # else it is a reference from an Init Combi - ignore
#                       print $longUsedProgOutFile "          Setlist [??] |$_|  \n"
                    }
              
                  }
               }
               print $UsedProgOutFile "\n";    # final LF

               
        }
        else {                    # Program is Unused
                $printline  = sprintf("%-7s : %-24s Unused",$ThisProgID, $ProgLine[0]);
#               print "$printline\n";
               print $unUsedProgOutFile "$printline\n";
               $ProgIsUsedHash{$ThisProgID} = "_";        #  record unused program with _
            }
            
     }
  } 
  close ($Patchfile);
 } else {
      die "Patch list (.CSV format) Filename Required as 2nd parameter";
 }

close ($UsedProgOutFile);
close ($longUsedProgOutFile);
close ($unUsedProgOutFile);






# --------------------------------------------------------------------------------------------------
# --------------------Combi Section ----------------------------------------------------------------
# --------------------------------------------------------------------------------------------------
# ---------------Output files for Combis ------------------

$UsedCombiOutFileName = $ARGV[0];
$UsedCombiOutFileName =~ s/\..*$/.usedcombi.txt/;  # force output filename to .usedcombi.txt
print "opening $UsedCombiOutFileName for writing\n";
open ($UsedCombiOutFile, '>'.$UsedCombiOutFileName);    

print $UsedCombiOutFile "Used Combi List for $ARGV[0] and $ARGV[1]\n";
print $UsedCombiOutFile "--------------------------------------------------------------\n";

$longUsedCombiOutFileName = $UsedCombiOutFileName;  
$longUsedCombiOutFileName =~ s/\..*$/.longusedcombi.txt/;  # force output filename to .longusedcombi.txt
print "opening $longUsedCombiOutFileName for writing\n";
open ($longUsedCombiOutFile, '>'.$longUsedCombiOutFileName);    
 
print $longUsedCombiOutFile "Long Format (with referencing Setlist Bank and Slot Names) Used Combi List for $ARGV[0] \n";
print $longUsedCombiOutFile "--------------------------------------------------------------\n";


$unUsedCombiOutFileName = $ARGV[0];
$unUsedCombiOutFileName =~ s/\..*$/.unusedcombi.txt/;  # force output filename to .unusedcombi.txt
print "opening $unUsedCombiOutFileName for writing\n";
open ($unUsedCombiOutFile, '>'.$unUsedCombiOutFileName);    
print $unUsedCombiOutFile "Unused Combi List for $ARGV[0] and $ARGV[1]\n";
print $unUsedCombiOutFile "--------------------------------------------------------------\n";





#  Read through patch list again and look at Combis for Setlist Usage
#     Look up  Combis in Setlist file
# 003/0, Abbey Road 01, , Combi, U-F000, Color 0, Font 0,  Transpose 0, HoldTime 6, Volume 127, Kbd_Track 0
#  [0]                  [1]     [2]   

if (defined($ARGV[1]))  {    # use second filename for the Patch list (.CSV format)
   print "Opening $ARGV[1] for reading\n";
   open (my $Patchfile, '<', $ARGV[1])
     or die "Could not open $ARGV[1]\n";

   while (my $work = <$Patchfile>) {
     chomp $work;
     $work =~ s/^\s+//;    # remove leading whitespace from work line
     
     if (index($work, ",Combi  ,") != -1) {    #is the patch list line referencing  a Combi? 
        $linecnt = $linecnt + 1;
        @ProgLine = split(',', $work);   # split patch list line on commas
        
        # format of line:  Acoustic Bass 2 SW1,Program,I-D097, ,,Bass/Synth Bass,A.Bass,,,
        #                        [0]              [1]     [2]
        #
        my $ThisProgID = $ProgLine[2];
#        print "$ThisProgID\n";
        if ($SetListUsageHash{$ThisProgID})  {    # it this Combi in the patch list found in Usage List hash? 
               my $SlotsUsing = $SetListUsageHash{$ThisProgID};    # get the string of Setlist data that use this program
               $printline  = sprintf("%-7s : %-24s UsedBy: ",$ThisProgID, $ProgLine[0]);   # print the start of the short format used list
#               print "$printline";
               print $UsedCombiOutFile "$printline";
               my $UseArrayLine = sprintf("%-7s *",$ThisProgID);   
               $CombiIsUsedHash{$ThisProgID} = "*";      #  record used program with *
               
               #  Now write out the long format with Using Setlist Names
               my(@SlotsUsingLine) = split(/,/, $SlotsUsing);    # Parse comma delimited items in the string into an array
               $printline  = sprintf("%-7s : %-24s",$ThisProgID, $ProgLine[0]);  # format the line to print describing the Combi
               print $longUsedCombiOutFile "$printline\n";

               foreach (@SlotsUsingLine)  { 
                   my $aslotid = $_;
                   $aslotid =~ s/^\s+|\s+$//g;  #trim lead/trail spaces   
#                   print "$aslotid\n";
#                   if (defined($SetlistHash{substr($_,0,3)})) {    # is it a known setlist from the patch list? 
                   if (defined($SetlistHash{$aslotid})) {    # is it a known setlist from the patch list? 
                      print $UsedCombiOutFile "$aslotid,";    # print only the setlist ID in short used file
                      print $longUsedCombiOutFile "          Setlist $SetlistHash{$aslotid} \n";    # print out the setlist bank and name
                   } else {                                               # else it is a reference from an Init Combi - ignore
                       print $longUsedCombiOutFile "          Setlist [??] |$_|  \n"
                   }
               }
               print $UsedCombiOutFile "\n";    # final LF

               
        }
        else {                    # Program is Unused
                $printline  = sprintf("%-7s : %-24s Unused",$ThisProgID, $ProgLine[0]);
#               print "$printline\n";
               print $unUsedCombiOutFile "$printline\n";
               $CombiIsUsedHash{$ThisProgID} = "_";        #  record unused program with _
            }
            
     }
  } 
  close ($Patchfile);
 } else {
      die "Patch list (.CSV format) Filename Required as 2nd parameter";
 }


close ($UsedCombiOutFile);
close ($longUsedCombiOutFile);
close ($unUsedCombiOutFile);

# --------------------------------------------------------------------------------------------------
# --------------------ASCII Section - Programs ---------------------------------------------------------------
# --------------------------------------------------------------------------------------------------


$AsciiOutFilename = $ARGV[0];      
$AsciiOutFilename =~ s/\..*$/.ascii.txt/;  # force output filename to .ascii.txt
print "opening $AsciiOutFilename for writing\n";
open ($AsciiOutFile, '>'.$AsciiOutFilename);    
print $AsciiOutFile "ASCII Chart of Program Useage for $ARGV[0]\n";
print $AsciiOutFile "Key:    \"*\" Prog referenced by a Combi or Setlist   \"_\" Non-Init Prog Unreferenced   \" \"  Init Program\n";
print $AsciiOutFile "--------------------------------------------------------------\n";

my $lastprefix = "";      
my $thisprefix = "";      
my $thisnumber = "";      
my $a;


my @banknames = ("I-A", "I-B", "I-C", "I-D", "I-E", "I-F", "U-A", "U-B", "U-C", "U-D", "U-E", "U-F", "U-G", "U-AA", "U-BB", "U-CC", "U-DD", "U-EE", "U-FF", "U-GG");
my @banktypes = ("EXi", "HD-1", "HD-1", "HD-1", "HD-1", "HD-1", "HD-1", "EXi", "EXi", "EXi", "EXi", "EXi", "EXi", "EXi", "HD-1", "HD-1", "EXi", "HD-1", "HD-1", "HD-1"); 


my $BankTypeFilename = $ARGV[2];
# $BankTypeFilename =~ s/_usage_output//;     # remove suffix to get base filename
$BankTypeFilename =~ s/\..*$/_banktype.txt/;  # force output filename to _banktype.txt
print "opening $BankTypeFilename for reading\n";

open(BankTypeFile1, "$BankTypeFilename") or die "Could not open $BankTypeFilename\n";
my %BankHash;
     while (my $Bankline=<BankTypeFile1>) {   
     chomp($Bankline);
   (my $Bankword1,my $Bankword2) = split /,/, $Bankline;  
          $BankHash{$Bankword1} = $Bankword2;   
    }
#    print Dumper(\%BankHash);
close(BankTypeFile1);

# foreach my $n (@banknames) {    
#   my $paddedID = sprintf("%-4s",$n);
#   print $AsciiOutFile "$paddedID | ";
# }
# print $AsciiOutFile "\n";
# foreach my $n (@banktypes) {    
#   my $paddedID = sprintf("%-4s",$n);
#   print $AsciiOutFile "$paddedID | ";
# }
# print $AsciiOutFile "\n\n";

print $AsciiOutFile "     0                               32                               64                               96                              127";
foreach my $n (@banknames) {    # Have to explicity iterate over all banks since "Ignore Empty Programs" cannot be disabled
#  print "$n  ";
  my $paddedID = sprintf("%-4s",$n);
  print $AsciiOutFile "\n$paddedID";

  
  for( $a = 0; $a < 128; $a = $a + 1 ) {   # Have to explicity iterate over all programs since "Ignore Empty Programs" cannot be disabled
     my $GenProgId = sprintf("%s%03d",$n,$a);  # generate the Program ID found in the hash
     unless (($a) % 32 ) {
           print $AsciiOutFile "|";
    }
     if ($ProgIsUsedHash{$GenProgId})  {    # it this program  found in Usage List hash? 
 #       print "Found $GenProgId [$ProgIsUsedHash{$GenProgId}]"
       print $AsciiOutFile "$ProgIsUsedHash{$GenProgId}";  #Print out * or _ indicating program is used or unreferenced
  }
  else {
       print $AsciiOutFile " ";      # Program not in ProgIsUsedHash - not in Patch List - must be Init Program
 #     print "Not Found $GenProgId"
  }
}
my $banktypestr = $BankHash{$n};
# print $banktypestr; 
print $AsciiOutFile "     $banktypestr";

}







# --------------------------------------------------------------------------------------------------
# --------------------ASCII Section - Combis ---------------------------------------------------------------
# --------------------------------------------------------------------------------------------------

# File still open from above   
print $AsciiOutFile "\n\n\nASCII Chart of Combi Useage for $ARGV[0]\n";
print $AsciiOutFile "Key:    \"*\" Combi referenced by a Setlist   \"_\" Non-Init Combi Unreferenced   \" \"  Init Combi\n";
print $AsciiOutFile "--------------------------------------------------------------\n";

$lastprefix = "";      
$thisprefix = "";      
$thisnumber = "";      


my @combibanknames = ("I-A", "I-B", "I-C", "I-D", "I-E", "I-F", "I-G", "U-A", "U-B", "U-C", "U-D", "U-E", "U-F", "U-G");
print $AsciiOutFile "     0                               32                               64                               96                              127";
foreach my $n (@combibanknames) {    # Have to explicity iterate over all banks since "Ignore Empty Programs" cannot be disabled
#  print "$n  ";
  my $paddedID = sprintf("%-4s",$n);
  print $AsciiOutFile "\n$paddedID";

  
  for( $a = 0; $a < 128; $a = $a + 1 ) {   # Have to explicity iterate over all combis 
     my $GenCombID = sprintf("%s%03d",$n,$a);  # generate the Combi ID found in the hash
     unless (($a) % 32 ) {
           print $AsciiOutFile "|";
    }
     if ($CombiIsUsedHash{$GenCombID})  {    # it this Combi found in Usage List hash? 
#       print "Found $GenCombID [$CombiIsUsedHash{$GenCombID}]\n";
       print $AsciiOutFile "$CombiIsUsedHash{$GenCombID}";  #Print out * or _ indicating program is used or unreferenced
  }
  else {
       print $AsciiOutFile " ";      # Program not in ProgIsUsedHash - not in Patch List - must be Init Program
#      print "Not Found $GenCombID\n";
  }
}
}


close($AsciiOutFile);










print "done\n";




exit;
